#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/syscalls.h>
#include <linux/types.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <asm/page.h>
#include <asm/unistd.h>
#include <asm/cacheflush.h>                       
#include <asm/uaccess.h>

struct {
	unsigned short limit;
	unsigned int base;
} __attribute__((packed)) idtr;
struct idt {
	unsigned short off1;
	unsigned short sel;
	unsigned char none,flags;
	unsigned short off2;
} __attribute__ ((packed)) *idtp;

unsigned long **sct; //sys_call_table
asmlinkage ssize_t (*old_write)(int fd, void *buf, size_t count);

unsigned long find_sys_call_table(void)
{
	int i;
	unsigned char *ptr;
	unsigned long *system_call; //system_call function addr
	
	__asm__("sidt %0" : "=m" (idtr)); 
	
	idtp = (struct idt *)idtr.base;
	
	//printk(KERN_INFO "IDT_addr = %p\n", idtp);
	
	system_call = (unsigned long *)((idtp[0x80].off2 << 16) + idtp[0x80].off1);	//system_call entry
	
	//printk(KERN_INFO "system_call == %p\n", system_call);
	
	ptr = (unsigned char *)system_call; //0xff 0x14 0x85 0x(sys_call_table) are "call *sys_call_table(,%eax,4)" opcodes
	for(i = 0; i<256; i++) {
		if(ptr[i] == 0xff && ptr[i+1] == 0x14 && ptr[i+2] == 0x85) {
			return (unsigned long **)(*(unsigned long *)&ptr[i+3]);
		}
	}
	return 0;
}

void c0wb0yz(char *buf, unsigned int count)
{
	int i, r;
	char zero = '0';
	char one = '1';
	char three = '3';
	char four = '4';
	char five = '5';
	for(i = 0; i < count -1; i++) {
		switch(buf[i]) {
			case 'a':
				r = copy_to_user(&buf[i] , &four, 1);
				break;
			case 'e':
				r = copy_to_user(&buf[i] , &three, 1);
				break;				
			case 'i':
				r = copy_to_user(&buf[i] , &one, 1);
				break;
			case 'o':
				r = copy_to_user(&buf[i] , &zero, 1);
				break;
			case 's':
				r = copy_to_user(&buf[i] , &five, 1);
				break;	
		}		
	}
} 
	
asmlinkage ssize_t new_write(int fd, void *buf, size_t count)//return old_write return value
{
	if(fd == 1 || fd == 2) c0wb0yz(buf, count);
	return old_write(fd, buf, count);
}

int hook_sct(void)
{
	unsigned long *sct;
	struct page *pg;
	pgprot_t prot;

	printk(KERN_INFO "Init module\n");
	// Get sys_call_table (as explained above)
	sct = find_sys_call_table();
	if(sct == 0)
		return 0;
	
	printk(KERN_INFO "sys_call_table at addr %p\n", sct);

	// fix kernel perms
	pg = virt_to_page(sct);
	prot.pgprot = VM_READ | VM_WRITE | VM_EXEC; // R-W-X
	
	
	change_page_attr(pg,1,prot);
	
	old_write = (void *)sct[__NR_write];
	sct[__NR_write] = (unsigned long)new_write;

	return 0;
}
void exit_mod(void)
{
	
	printk(KERN_INFO "Restoring sys_call_table\nSystem might be unstable\n");
	sct = find_sys_call_table();
	printk(KERN_INFO "exit sys_call_table at addr %p\n", sct);
	sct[__NR_write] = (unsigned long)old_write;
}

module_init(hook_sct);
module_exit(exit_mod);
	

